#include "terrainmap.h"
#include <stdlib.h>

// ______________________________________
// Implementation of CTerrainSector Class

CTerrainSector::CTerrainSector(	unsigned short	uSize,
								float			fHillMin,
								float			fHillMax,
								float			fMaxHillHeight,
								unsigned short	uNumHills,
								unsigned short	uFlattening,
								bool			bIsland,
								unsigned int	uSeed,
								GLuint			uPlainDL,
								unsigned int	uTextureID )
{
	m_Thread = NULL;
	m_HandleMe = false;
	m_IsCreating = false;
	m_IsCreated = false;
	m_Sector = NULL;
	m_TerrainDL = NULL;
	m_PlainDL = uPlainDL;
	m_TextureID = uTextureID;
	m_Size = uSize;
	m_HillMin = fHillMin;
	m_HillMax = fHillMax;
	m_MaxHillHeight = fMaxHillHeight;
	m_NumHills = uNumHills;
	m_Flattening = uFlattening;
	m_Island = bIsland;
	m_Seed = uSeed;
}
CTerrainSector::~CTerrainSector(void)
{
	if (m_IsCreated)
	{
		// delete display list and terrain data
		glDeleteLists(m_TerrainDL, 1);
		delete m_Sector;
	}
	else if(m_IsCreating)
	{
		CloseHandle( m_Thread );
		delete m_Sector;
	}
}

void CTerrainSector::create(bool &isThreading)
{
	// When the thread is done, close the handle and deal with new data
	if(m_HandleMe)
	{
		CloseHandle( m_Thread );
		m_HandleMe = false;

		// Create Terrain Display List
		m_TerrainDL = glGenLists(1);
		glNewList(m_TerrainDL, GL_COMPILE);
		glColor3f(0.55f,0.42f,0.24f);
		glEnable( GL_TEXTURE_2D );
		C3dVector normal;
		glBindTexture( GL_TEXTURE_2D, m_TextureID);
		glTexEnvf( GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE );
		for( int z = 1; z < m_Sector->GetSize(); z++ )
		{
			glBegin(GL_TRIANGLE_STRIP);
			for( int x = 0; x < m_Sector->GetSize(); x++ )
			{
				normal = m_Sector->GetNormal(x, z);

				glNormal3f(	normal.getX(), normal.getY(), normal.getZ() );
				glTexCoord2d(x%2,1); glVertex3f(x, m_Sector->GetCell(x, z), -z);

				normal = m_Sector->GetNormal(x, z-1);
				glNormal3f(	normal.getX(), normal.getY(), normal.getZ() );
				glTexCoord2d(x%2,0); glVertex3f(x, m_Sector->GetCell(x, z-1), -(z-1));
			}
			glEnd();
		}
		glEndList();

		m_IsCreated = true;
		m_IsCreating = false;
		isThreading = false;
		return;
	}

	if(m_IsCreated || m_IsCreating || isThreading) return;
	
	// --- Create a new Terrain ---
	isThreading = true;
	m_IsCreating = true;

	DWORD dwThreadId;
    HANDLE hThread;

	m_Sector = new CHillTerrain(	m_Size, m_HillMin, m_HillMax, m_MaxHillHeight,
									m_NumHills, m_Flattening, m_Island, m_Seed );

	m_Thread = CreateThread(
		NULL,											// default security attributes 
		NULL,											// use default stack size  
		( LPTHREAD_START_ROUTINE )ThreadedGenerator,	// thread function 
		this,											// argument to thread function 
		CREATE_SUSPENDED,								// creation flag to first set priority
		NULL);											// don't need thread identifier
	
	// make sure terrain creation doesn't interfere with game frame rate too much
	//SetThreadPriority(m_Thread, THREAD_PRIORITY_BELOW_NORMAL);
	SetThreadPriority(m_Thread, THREAD_PRIORITY_LOWEST);
	ResumeThread(m_Thread);
}

void CTerrainSector::ThreadedGenerator(CTerrainSector *pointer)
{
	pointer->m_Sector->Generate();

	// thread done; request to have handle closed and signal to process terrain-data
	pointer->m_HandleMe = true;
}

void CTerrainSector::destroy(bool &isThreading)
{
	if(!m_IsCreated) return;

	// this may be confusing, but the reason I do it is to make
	// sure that WHEN the thread is done, it is handled properly (and reliably).
	// a subsequent call to destroy() will then work as expected
	if(m_IsCreating)
	{
		create(isThreading);
		return;
	}

	// delete display list and terrain data
	glDeleteLists(m_TerrainDL, 1);
	delete m_Sector;
	m_IsCreated = false;
}

inline void CTerrainSector::draw(void)
{
	if(m_IsCreated) glCallList(m_TerrainDL);
	else glCallList(m_PlainDL);
}

double CTerrainSector::CollisionPoint(float x, float z)
// NOTE: the z value is flipped (so it is treated like a y-component)
// PRE: x and z cannot be >= m_Size-1
{
	if(!m_IsCreated) return 0.0;

	// Determine heights at 4 vertices
	double v1 = m_Sector->GetCell(x, z);
	double v2 = m_Sector->GetCell(x+1, z);
	double v3 = m_Sector->GetCell(x+1, z+1);
	double v4 = m_Sector->GetCell(x, z+1);

	// Determine weights
	double x_weight = x - ((int) x);
	double z_weight = z - ((int) z);

	// Return aprox interpolated height
	return (((v3-v4)*x_weight + v4)-((v2-v1)*x_weight + v1))*z_weight + ((v2-v1)*x_weight + v1);
}



// ___________________________________
// Implementation of CTerrainMap Class

CTerrainMap::CTerrainMap(unsigned short visibility_detail, unsigned int random_seed )
{
	m_isThreading = false;
	srand( random_seed );

	if(visibility_detail > 20) visibility_detail = 20;
	
	// the boundaries for terrain generation
	m_LowBound2 = (++visibility_detail)*SECTOR;
	m_HighBound2 = (visibility_detail+1)*SECTOR;
	
	// the boundaries for terrain deletion
	m_LowBound1 = 1.35*m_LowBound2;
	m_HighBound1 = 1.35*m_HighBound2;

	// fog visibility for skybox
	GLfloat fogDensity = 0.0005f/((float) visibility_detail);

	// Fog Code (Modified) from NeHe OpenGL Tutorial 16
	GLfloat fogColor[4]= {0.55f, 0.42f, 0.24f, 1.0f};	// Fog Color
	glClearColor(0.55f,0.42f,0.24f,1.0f);				// Clear to Color of Fog
	glFogi (GL_FOG_MODE, GL_LINEAR);					// Fog fade function
	glFogfv(GL_FOG_COLOR, fogColor);					// Fog Color
	glFogf(GL_FOG_DENSITY, fogDensity);					// Fog Density (for Skybox)
	glHint(GL_FOG_HINT, GL_NICEST);						// Fog Hint Value
	glFogf(GL_FOG_START, m_LowBound2/2);				// Fog Start Depth (for terrain)
	glFogf(GL_FOG_END, m_LowBound2);					// Fog End Depth (for terrain)
	glEnable(GL_FOG);									// Enables GL_FOG

	// prepare textures
	m_Desert_Texture.LoadFromFile("desert.bmp");
	m_Tex1.LoadFromFile("back.bmp");
	m_Tex2.LoadFromFile("front.bmp");
	m_Tex3.LoadFromFile("top.bmp");
	m_Tex4.LoadFromFile("left.bmp");
	m_Tex5.LoadFromFile("right.bmp");

	GLuint dList = m_dList = glGenLists(2);
	
	// Create SkyBox Display List
	glNewList(m_dList, GL_COMPILE);
		glEnable( GL_TEXTURE_2D );
		glPushMatrix();
			glRotatef(45, 0, 1, 0); // rotate so it's harder for humans to see "box"
			CreateSkyBox(0, 0, 0, 5000, 5000, 5000);
		glPopMatrix();
	glEndList();

	// Create Plain Display List
	glNewList(m_dList+1, GL_COMPILE);
		glDisable( GL_TEXTURE_2D );
		glColor3f(0.55f,0.42f,0.24f);
		glBegin(GL_TRIANGLE_STRIP);
			glNormal3f(	0, 1, 0 ); glVertex3f(0, 0, 0);
			glNormal3f(	0, 1, 0 ); glVertex3f(SECTOR, 0, 0);
			glNormal3f(	0, 1, 0 ); glVertex3f(0, 0, -SECTOR);
			glNormal3f(	0, 1, 0 ); glVertex3f(SECTOR, 0, -SECTOR);
		glEnd();
	glEndList();

	// initialize grid (terrain create map)
	int random = rand();
	for (int x = 0; x < 20; x++)
	{
		for (int y = 0; y < 20; y++)
		{
			dList++;
			if ( (x == 0) || (x == 19) || (y == 0) || (y == 19) )
                m_grid[x][y] = new CTerrainSector(	SECTOR+1, 5, 20, 100, 500, 1, false,
													random+x+y, m_dList+1, m_Desert_Texture.GetID() );
			else
				m_grid[x][y] = new CTerrainSector(	SECTOR+1, 2, 18, 12, 200, 1, false,
													random+x+y, m_dList+1, m_Desert_Texture.GetID() );
		}
	}
}

CTerrainMap::~CTerrainMap(void)
{
	// delete terrains
	for (int x = 0; x < 20; x++)
	{
		for (int y = 0; y < 20; y++)
		{
			delete m_grid[x][y];
		}
	}

	// delete display lists
	glDeleteLists(m_dList, 2);
}

void CTerrainMap::draw(int camera_X, int camera_Z)
{
	// Convert camera coordinates to map coordinates
	camera_X += 1270;
	camera_Z = -camera_Z + 1270;
	if (camera_X > 2540) camera_X = 2540;
	if (camera_Z > 2540) camera_Z = 2540;
	if (camera_X < 0) camera_X = 0;
	if (camera_Z < 0) camera_Z = 0;

	// Make the terrain look like sand
	GLfloat shineColor[] = { 0.0, 0.0, 0.0, 1.0 };
	glMaterialfv(GL_FRONT, GL_SPECULAR, shineColor);
	glMaterialf(GL_FRONT, GL_SHININESS, 0);

	// determine which sectors to be hill terrain and which to be simple plains
	for (int x = 0; x < 20; x++)
	{
		for (int z = 0; z < 20; z++)
		{
			if(	(camera_X >= ( (x*SECTOR)-m_LowBound1 )) && (camera_X < ( (x*SECTOR)+m_HighBound1 )) &&
				(camera_Z >= ( (z*SECTOR)-m_LowBound1 )) && (camera_Z < ( (z*SECTOR)+m_HighBound1 )) )
			{
				if(	(camera_X >= ( (x*SECTOR)-m_LowBound2 )) && (camera_X < ( (x*SECTOR)+m_HighBound2 )) &&
					(camera_Z >= ( (z*SECTOR)-m_LowBound2 )) && (camera_Z < ( (z*SECTOR)+m_HighBound2 )) )
				{
					m_grid[x][z]->create(m_isThreading);
				}
				// else don't delete yet, cuz we may go back
			}
			else
			{
				m_grid[x][z]->destroy(m_isThreading);
			}
		}
	}

	// draw skybox
	glCallList(m_dList);
	
	// draw map
	glPushMatrix();
		glTranslatef(-1270, 0, 1270);
		for (int x = 0; x < 20; x++)
		{
			for (int z = 0; z < 20; z++)
			{
				m_grid[x][z]->draw();
				glTranslatef(0, 0, -127);
			}
			glTranslatef(127, 0, 2540);
		}
	glPopMatrix();
}

double CTerrainMap::CollisionPoint(float x, float z)
{
	// check to see if point is over terrain
	// i have used 1269, instead of 1270 to avoid another if-statement (to prevent seg-
	// fault when accessing m_grid). Tanks and debris should NEVER reach there anyway!
	if ( (x > 1269) || (x < -1269) || (z < -1269) || (z > 1269) ) return -1;
	
	x += 1270;
	z = -z + 1270;

	// Break float into quotient and remainder
	int x_sector = x/SECTOR; x -= x_sector*SECTOR;
	int z_sector = z/SECTOR; z -= z_sector*SECTOR;

	return m_grid[x_sector][z_sector]->CollisionPoint(x, z);
}

void CTerrainMap::VehicleCollision(float x1, float z1, float x2, float z2, float x3, float z3, float x4, float z4, float& pitch, float& roll)
{
	double p1 = CollisionPoint(x1, z1);
	double p2 = CollisionPoint(x2, z2);
	double p3 = CollisionPoint(x3, z3);
	double p4 = CollisionPoint(x4, z4);
	if ( (p1 < 0) || (p2 < 0) || (p3 < 0) || (p4 < 0) ) return;

	pitch = atan((p1-p2)/(sqrt((x1-x2)*(x1-x2) + (z1-z2)*(z1-z2))));
	roll = atan((p4-p3)/(sqrt((x4-x3)*(x4-x3) + (z4-z3)*(z4-z3))));
}

void CTerrainMap::CreateSkyBox(float x, float y, float z, float width, float height, float length)
{
	// Turn on texture mapping if it's not already
	glEnable(GL_TEXTURE_2D);

	// Clear, bright textures
	glColor3f(1, 1, 1);
	glDisable(GL_LIGHTING);

	// fog is linear for map but exponential for skybox (looks better)
	glFogi (GL_FOG_MODE, GL_EXP);

	// Bind the BACK texture of the sky map to the BACK side of the cube
	m_Tex1.SetActive();
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);

	// This centers the sky box around (x, y, z)
	x = x - width  / 2;
	y = y - height / 2;
	z = z - length / 2;

	// Start drawing the side as a QUAD
	glBegin(GL_QUADS);		
		
		// Assign the texture coordinates and vertices for the BACK Side
		glTexCoord2f(1.0f, 0.0f); glVertex3f(x + width, y,			z);
		glTexCoord2f(1.0f, 1.0f); glVertex3f(x + width, y + height, z);
		glTexCoord2f(0.0f, 1.0f); glVertex3f(x,			y + height, z);
		glTexCoord2f(0.0f, 0.0f); glVertex3f(x,			y,			z);
		
	glEnd();

	// Bind the FRONT texture of the sky map to the FRONT side of the box
	m_Tex2.SetActive();
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);

	// Start drawing the side as a QUAD
	glBegin(GL_QUADS);	
	
		// Assign the texture coordinates and vertices for the FRONT Side
		glTexCoord2f(1.0f, 0.0f); glVertex3f(x,			y,			z + length);
		glTexCoord2f(1.0f, 1.0f); glVertex3f(x,			y + height, z + length);
		glTexCoord2f(0.0f, 1.0f); glVertex3f(x + width, y + height, z + length); 
		glTexCoord2f(0.0f, 0.0f); glVertex3f(x + width, y,			z + length);
	glEnd();

	// Bind the TOP texture of the sky map to the TOP side of the box
	m_Tex3.SetActive();
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);

	// Start drawing the side as a QUAD
	glBegin(GL_QUADS);		
		
		// Assign the texture coordinates and vertices for the TOP Side
		glTexCoord2f(0.0f, 1.0f); glVertex3f(x + width, y + height, z);
		glTexCoord2f(0.0f, 0.0f); glVertex3f(x + width, y + height, z + length);
		glTexCoord2f(1.0f, 0.0f); glVertex3f(x,			y + height,	z + length);
		glTexCoord2f(1.0f, 1.0f); glVertex3f(x,			y + height,	z);
		
	glEnd();

	// Bind the LEFT texture of the sky map to the LEFT side of the box
	m_Tex4.SetActive();
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);

	// Start drawing the side as a QUAD
	glBegin(GL_QUADS);		
		
		// Assign the texture coordinates and vertices for the LEFT Side
		glTexCoord2f(1.0f, 1.0f); glVertex3f(x,			y + height,	z);
		glTexCoord2f(0.0f, 1.0f); glVertex3f(x,			y + height,	z + length);
		glTexCoord2f(0.0f, 0.0f); glVertex3f(x,			y,			z + length);
		glTexCoord2f(1.0f, 0.0f); glVertex3f(x,			y,			z);		
		
	glEnd();

	// Bind the RIGHT texture of the sky map to the RIGHT side of the box
	m_Tex5.SetActive();
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);

	// Start drawing the side as a QUAD
	glBegin(GL_QUADS);		

		// Assign the texture coordinates and vertices for the RIGHT Side
		glTexCoord2f(0.0f, 0.0f); glVertex3f(x + width, y,			z);
		glTexCoord2f(1.0f, 0.0f); glVertex3f(x + width, y,			z + length);
		glTexCoord2f(1.0f, 1.0f); glVertex3f(x + width, y + height,	z + length); 
		glTexCoord2f(0.0f, 1.0f); glVertex3f(x + width, y + height,	z);
	glEnd();

	// Return settings back
	glEnable(GL_LIGHTING);
	glFogi (GL_FOG_MODE, GL_LINEAR);
}

inline bool CTerrainMap::PendingThread(void)
{
	return m_isThreading;
}

